.carousel-arrow-container {
  border-bottom: 1px solid @borders;
}

.carousel-arrow,
.carousel-inner-arrow {
  border-width: 20px; /* ARROW_SIZE */
  border-style: solid;
  border-color: transparent;
}

.carousel-arrow {
  border-bottom-color: @borders;
  margin-bottom: -1px;
  animation-duration: 200ms;
  animation-timing-function: ease-in-out;
  animation-fill-mode: forwards;
}

.carousel-inner-arrow {
  border-bottom-color: @theme_bg_color;
  margin-bottom: -2px;
}

.carousel-item {
  background: transparent;
  box-shadow: none;
  border: none;
  color: @theme_fg_color;
}
