frame.desktop-preview {
  min-height: 10px;
  padding: 0 4px;
  background-color: black;
}

frame.desktop-preview image {
  color: white;
}

frame.desktop-preview label {
  color: white;
  font-weight: bold;
  font-size: 6px;
}

frame.lock-screen-preview {
  border: solid rgba(0, 0, 0, 0.33);
  border-width: 10px 0 0 0;
}

frame.lock-screen-preview label {
  color: white;
  font-weight: bold;
  text-shadow: 0 1px 2px rgba(0, 0, 0, 0.6);
  font-size: 1.2em;
}

image.slideshow-icon {
  color: white;
  -gtk-icon-shadow: 0 1px 2px rgba(0, 0, 0, 0.33);
}
